#ifndef DALI_INTEGRATION_TOUCH_EVENT_COMBINER_H
#define DALI_INTEGRATION_TOUCH_EVENT_COMBINER_H

/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// INTERNAL INCLUDES
#include <dali/public-api/common/vector-wrapper.h>
#include <dali/public-api/math/vector2.h>
#include <dali/integration-api/events/point.h>

namespace Dali
{

namespace Integration
{

struct TouchEvent;
struct HoverEvent;

/**
 * Dali::Integration::TouchEventCombiner is a utility class, an instance of which, should be created
 * upon initialisation.  It accepts single Point(s) containing information about a touch area and
 * creates a TouchEvent and/or HoverEvent combining the latest event's information with previous Point(s).
 *
 * The created TouchEvent and/or HoverEvent can then be sent to the Dali Core as indicated by the GetNextTouchEvent()
 * method.
 *
 * The TouchEventCombiner ensures that the following rules are also adhered to:
 * - If two Down events are accidentally received, then the second one is ignored.
 * - Motion events are not processed if a Down event does not precede it.
 * - Motion event throttling is carried out to satisfy the minimum distance and time delta required.
 * - If an interrupted event is received, then any stored Point history is cleared.
 */
class DALI_CORE_API TouchEventCombiner
{
public:

  // Enumerations

  enum EventDispatchType
  {
    DispatchTouch,      ///< The touch event should be dispatched.
    DispatchHover,      ///< The hover event should be dispatched.
    DispatchBoth,       ///< Both touch event and hover event should be dispatched.
    DispatchNone        ///< Neither touch event nor hover event should be dispatched.
  };

  /**
   * Default Constructor.
   * @note The default minimum motion time is 1 ms between motion events and the default behaviour
   *       is to throttle X and Y movement by 1.
   */
  TouchEventCombiner();

  /**
   * Construction with parameters.
   * @param[in]  minMotionTime       The minimum time (in ms) that should occur between motion events.
   * @param[in]  minMotionXDistance  The minimum distance a finger has to be moved between horizontal motion events.
   * @param[in]  minMotionYDistance  The minimum distance a finger has to be moved between vertical motion events.
   * @note Will assert if any of the parameters is negative.
   */
  TouchEventCombiner( uint32_t minMotionTime, float minMotionXDistance, float minMotionYDistance );

  /**
   * Construction with parameters.
   * @param[in]  minMotionTime      The minimum time (in ms) that should occur between motion events.
   * @param[in]  minMotionDistance  A Vector2 representing the minimum distance a finger has to be moved between horizontal and vertical motion events.
   * @note Will assert if any of the parameters is negative.
   */
  TouchEventCombiner( uint32_t minMotionTime, Vector2 minMotionDistance );

  /**
   * Non virtual destructor
   */
  ~TouchEventCombiner();

public:

  /**
   * Allows the caller to pass in a point which is processed and the TouchEvent and/or HoverEvent is appropriately filled with the new,
   * and previously stored Point information.
   *
   * @note If the thresholds set have not been passed, then false is returned and the TouchEvent and/or HoverEvent should not be sent
   * to Dali Core.
   *
   * @param[in]   point         The new Point.
   * @param[in]   time          The time the event occurred.
   * @param[out]  touchEvent    This is populated with the correct Point(s) and time information.
   * @param[out]  hoverEvent    This is populated with the correct Point(s) and time information.
   *
   * @return true if the point is beyond the different thresholds set thus, should be sent to core, false otherwise.
   */
  EventDispatchType GetNextTouchEvent( const Point& point, uint32_t time, TouchEvent& touchEvent, HoverEvent& hoverEvent );

  /**
   * Sets the minimum time (in ms) that should occur between motion events.
   * @param[in]  minTime  Minimum time between motion events.
   */
  void SetMinimumMotionTimeThreshold( uint32_t minTime );

  /**
   * Sets the minimum distance a finger has to be moved (both X and Y) between motion events.
   * @param[in]  minDistance  The minimum distance between motion events.
   * @note Will assert if parameter is negative.
   */
  void SetMinimumMotionDistanceThreshold( float minDistance );

  /**
   * Sets the minimum distance a finger has to be moved between motion events.
   * @param[in]  minXDistance  The minimum X distance between motion events.
   * @param[in]  minYDistance  The minimum Y distance between motion events.
   * @note Use this method if the X and Y threshold required is different.
   * @note Will assert if any of the parameters is negative.
   */
  void SetMinimumMotionDistanceThreshold( float minXDistance, float minYDistance );

  /**
   * Sets the minimum distance a finger has to be moved between motion events.
   * @param[in]  minDistance  A Vector2 representing the minimum X and Y thresholds.
   * @note Use this method if the X and Y threshold required is different.
   * @note Will assert if any of the parameters is negative.
   */
  void SetMinimumMotionDistanceThreshold( Vector2 minDistance );

  /**
   * Retrieves the minimum motion time threshold.
   * @return the minimum time threshold.
   */
  unsigned long GetMinimumMotionTimeThreshold() const;

  /**
   * Gets the minimum distance a finger has to be moved (both X and Y) between motion events.
   * @return A Vector2 representing the x and y thresholds.
   */
  Vector2 GetMinimumMotionDistanceThreshold() const;

  /**
   * This resets any information contained by the TouchEventCombiner.
   * This may be required if some platform event has occurred which makes it necessary to reset any
   * Point information that the combiner may store.
   */
  void Reset();

private:

  struct PointInfo;
  typedef std::vector< PointInfo > PointInfoContainer;
  PointInfoContainer mPressedPoints; ///< A container of touched point and time.
  PointInfoContainer mHoveredPoints; ///< A container of hovered point and time.

  uint32_t mMinMotionTime; ///< The minimum time that should elapse before considering a new motion event.
  Vector2 mMinMotionDistance; ///< The minimum distance in the X and Y direction before considering a new motion event.
};

} // namespace Integration

} // namespace Dali

#endif // DALI_INTEGRATION_TOUCH_EVENT_COMBINER_H
